(function() {
  var ResourcesParser, debug;

  debug = require('debug')('adb:apkreader:parser:resources');

  ResourcesParser = (function() {
    var ChunkType, NodeType, StringFlags, TypedValue;

    NodeType = {
      ELEMENT_NODE: 1,
      ATTRIBUTE_NODE: 2,
      CDATA_SECTION_NODE: 4
    };


    AndroidConstants = {
      RESOURCE_FILE: "resources.arsc",
      MANIFEST_FILE: "AndroidManifest.xml",
      DEX_FIL: "classes.dex",
      RES_PREFIX: "res/",
      ASSETS_PREFIX: "assets/",
      LIB_PREFIX: "lib/",
      META_PREFIX: "META-INF/",
      ARCH_ARMEABI: "",
      ATTR_ID_START: 0x01010000,
      SYS_STYLE_ID_START: 0x01030000,
      SYS_STYLE_ID_END: 0x01031000
    }

    ChunkType = {
      NULL: 0x0000,
      STRING_POOL: 0x0001,
      TABLE: 0x0002,
      XML: 0x0003,
      XML_FIRST_CHUNK: 0x0100,
      XML_START_NAMESPACE: 0x0100,
      XML_END_NAMESPACE: 0x0101,
      XML_START_ELEMENT: 0x0102,
      XML_END_ELEMENT: 0x0103,
      XML_CDATA: 0x0104,
      XML_LAST_CHUNK: 0x017f,
      XML_RESOURCE_MAP: 0x0180,
      TABLE_PACKAGE: 0x0200,
      TABLE_TYPE: 0x0201,
      TABLE_TYPE_SPEC: 0x0202
    };

    StringFlags = {
      SORTED: 1 << 0,
      UTF8: 1 << 8
    };

    TypedValue = {
      COMPLEX_MANTISSA_MASK: 0x00ffffff,
      COMPLEX_MANTISSA_SHIFT: 0x00000008,
      COMPLEX_RADIX_0p23: 0x00000003,
      COMPLEX_RADIX_16p7: 0x00000001,
      COMPLEX_RADIX_23p0: 0x00000000,
      COMPLEX_RADIX_8p15: 0x00000002,
      COMPLEX_RADIX_MASK: 0x00000003,
      COMPLEX_RADIX_SHIFT: 0x00000004,
      COMPLEX_UNIT_DIP: 0x00000001,
      COMPLEX_UNIT_FRACTION: 0x00000000,
      COMPLEX_UNIT_FRACTION_PARENT: 0x00000001,
      COMPLEX_UNIT_IN: 0x00000004,
      COMPLEX_UNIT_MASK: 0x0000000f,
      COMPLEX_UNIT_MM: 0x00000005,
      COMPLEX_UNIT_PT: 0x00000003,
      COMPLEX_UNIT_PX: 0x00000000,
      COMPLEX_UNIT_SHIFT: 0x00000000,
      COMPLEX_UNIT_SP: 0x00000002,
      DENSITY_DEFAULT: 0x00000000,
      DENSITY_NONE: 0x0000ffff,
      TYPE_ATTRIBUTE: 0x00000002,
      TYPE_DIMENSION: 0x00000005,
      TYPE_FIRST_COLOR_INT: 0x0000001c,
      TYPE_FIRST_INT: 0x00000010,
      TYPE_FLOAT: 0x00000004,
      TYPE_FRACTION: 0x00000006,
      TYPE_INT_BOOLEAN: 0x00000012,
      TYPE_INT_COLOR_ARGB4: 0x0000001e,
      TYPE_INT_COLOR_ARGB8: 0x0000001c,
      TYPE_INT_COLOR_RGB4: 0x0000001f,
      TYPE_INT_COLOR_RGB8: 0x0000001d,
      TYPE_INT_DEC: 0x00000010,
      TYPE_INT_HEX: 0x00000011,
      TYPE_LAST_COLOR_INT: 0x0000001f,
      TYPE_LAST_INT: 0x0000001f,
      TYPE_NULL: 0x00000000,
      TYPE_REFERENCE: 0x00000001,
      TYPE_STRING: 0x00000003
    };


    ResValue_ResType = {
        // Contains no data.
       NULL: 0x00,
        // The 'data' holds a ResTable_ref, a reference to another resource
        // table entry.
       REFERENCE: 0x01,
        // The 'data' holds an attribute resource identifier.
       ATTRIBUTE: 0x02,
        // The 'data' holds an index into the containing resource table's
        // global value string pool.
       STRING: 0x03,
        // The 'data' holds a single-precision floating point number.
       FLOAT: 0x04,
        // The 'data' holds a complex number encoding a dimension value,
        // such as "100in".
       DIMENSION: 0x05,
        // The 'data' holds a complex number encoding a fraction of a
        // container.
       FRACTION: 0x06,

        // Beginning of integer flavors...
       FIRST_INT: 0x10,

        // The 'data' is a raw integer value of the form n..n.
       INT_DEC: 0x10,
        // The 'data' is a raw integer value of the form 0xn..n.
       INT_HEX: 0x11,
        // The 'data' is either 0 or 1, for input "false" or "true" respectively.
       INT_BOOLEAN: 0x12,

        // Beginning of color integer flavors...
       FIRST_COLOR_INT: 0x1c,

        // The 'data' is a raw integer value of the form #aarrggbb.
       INT_COLOR_ARGB8: 0x1c,
        // The 'data' is a raw integer value of the form #rrggbb.
       INT_COLOR_RGB8: 0x1d,
        // The 'data' is a raw integer value of the form #argb.
       INT_COLOR_ARGB4: 0x1e,
        // The 'data' is a raw integer value of the form #rgb.
       INT_COLOR_RGB4: 0x1f,

        // ...end of integer flavors.
       LAST_COLOR_INT: 0x1f,

        // ...end of integer flavors.
       LAST_INT: 0x1f
    }


    function ResourcesParser(buffer) {
      this.buffer = buffer;
      this.cursor = 0;
      this.strings = [];
      this.resources = [];
      this.document = null;
      this.parent = null;
      this.stack = [];
      this.locales = [];
      this.resourceTable = {};
    }

    ResourcesParser.prototype.readU8 = function() {
      var val;
      val = this.buffer[this.cursor];
      this.cursor += 1;
      return val;
    };

    ResourcesParser.prototype.readU16 = function() {
      var val;
      val = this.buffer.readUInt16LE(this.cursor);
      this.cursor += 2;
      return val;
    };

    ResourcesParser.prototype.readS32 = function() {
      var val;
      val = this.buffer.readInt32LE(this.cursor);
      this.cursor += 4;
      return val;
    };

    ResourcesParser.prototype.readU32 = function() {
      var val;
      val = this.buffer.readUInt32LE(this.cursor);
      this.cursor += 4;
      return val;
    };

    ResourcesParser.prototype.readLength8 = function() {
      var len = 0;
      var i = this.readU8();
      var j = 0;
      if ( (i & 0x80) !=0) {
        len = len | ( (i & 0x7f) << 8 );
        j = this.readU8();
        len += j;
      } else {
        len = i;
      }
      return len;
    };

    ResourcesParser.prototype.readLength16 = function() {
      var len;
      len = this.readU16();
      if (len & 0x8000) {
        len = (len & 0x7fff) << 15;
        len += this.readU16();
      }
      return len;
    };

    /**
     * method to read resource value RGB/ARGB type.
     */
    ResourcesParser.prototype.readRGBs = function(strLen) {
        var l = this.readU32();
        return this.getRGBs(l, strLen);
    }
    ResourcesParser.prototype.getRGBs = function(data, strLen) {
        var l = data;
        var sb = "";
        for (var i = strLen / 2 - 1; i >= 0; i--) {
            sb=sb + ((l >> i * 8) & 0xff).toString(16);
        }
        return sb;
    }

    /**
     * read res value, convert from different types to string.
     */
    ResourcesParser.prototype.readResValue = function(type, resourceTable, locale) {
        var size = type.buffer.readUInt16LE(type.cursor);
        type.cursor += 2;
        var res0 = type.buffer[type.cursor];
        type.cursor += 1;
        var dataType = type.buffer[type.cursor];
        type.cursor += 1;
        var resValue = {
          size: size,
          res0: res0,
          dataType: dataType
        }

        // console.log("readResValue %d %d %s", resValue.size, resValue.res0, resValue.dataType.toString(16));
        var data = type.buffer.readUInt32LE(type.cursor);
        switch (resValue.dataType) {
            case ResValue_ResType.INT_DEC:
            case ResValue_ResType.INT_HEX:
                resValue.value=data.toString();
                break;
            case ResValue_ResType.STRING:
                var strRef = data;
                if (strRef >= 0) {
                    resValue.value=type.stringPool[strRef];
                }
                break;
            case ResValue_ResType.REFERENCE:
                var resourceId = data;
                resValue.value = this.getResourceById(resourceId, resourceTable, locale);
                break;
            case ResValue_ResType.INT_BOOLEAN:
                resValue.value = (data) != 0 ? true : false;
                break;
            case ResValue_ResType.NULL:
                resValue.value = "";
                break;
            case ResValue_ResType.INT_COLOR_RGB8:
            case ResValue_ResType.INT_COLOR_RGB4:
                resValue.value = this.getRGBs(data, 6);
                break;
            case ResValue_ResType.INT_COLOR_ARGB8:
            case ResValue_ResType.INT_COLOR_ARGB4:
                resValue.value = this.getRGBs(data, 8);
                break;
            case ResValue_ResType.DIMENSION:
                resValue.value = this.getDimension(data);
                break;
            case ResValue_ResType.FRACTION:
                resValue.value = this.getFraction(data);
                break;
            default:
                resValue.value = data;
        }
        return resValue.value;
    }

    ResourcesParser.prototype.readDimension = function() {
      var value = this.readU32();
      this.getDimension(value);
    }
    ResourcesParser.prototype.getDimension = function(data) {
      var dimension, unit, value;
      dimension = {
        value: null,
        unit: null,
        rawUnit: null
      };
      value = data;
      unit = dimension.value & 0xff;
      dimension.value = value >> 8;
      dimension.rawUnit = unit;
      switch (unit) {
        case TypedValue.COMPLEX_UNIT_MM:
          dimension.unit = 'mm';
          break;
        case TypedValue.COMPLEX_UNIT_PX:
          dimension.unit = 'px';
          break;
        case TypedValue.COMPLEX_UNIT_DIP:
          dimension.unit = 'dp';
          break;
        case TypedValue.COMPLEX_UNIT_SP:
          dimension.unit = 'sp';
          break;
        case TypedValue.COMPLEX_UNIT_PT:
          dimension.unit = 'pt';
          break;
        case TypedValue.COMPLEX_UNIT_IN:
          dimension.unit = 'in';
      }
      return dimension;
    };

    ResourcesParser.prototype.readFraction = function() {
      var value = this.readU32();
      getFraction(value);
    }
    ResourcesParser.prototype.getFraction = function(data) {
      var fraction, type, value;
      fraction = {
        value: null,
        type: null,
        rawType: null
      };
      value = data;
      type = value & 0xf;
      fraction.value = this.convertIntToFloat(value >> 4);
      fraction.rawType = type;
      switch (type) {
        case TypedValue.COMPLEX_UNIT_FRACTION:
          fraction.type = '%';
          break;
        case TypedValue.COMPLEX_UNIT_FRACTION_PARENT:
          fraction.type = '%p';
      }
      return fraction;
    };

    ResourcesParser.prototype.readHex24 = function() {
      return (this.readU32() & 0xffffff).toString(16);
    };

    ResourcesParser.prototype.readHex32 = function() {
      return this.readU32().toString(16);
    };

    ResourcesParser.prototype.readTypedValue = function() {
      var dataType, diff, end, id, ref, size, start, type, typedValue, zero;
      typedValue = {
        value: null,
        type: null,
        rawType: null
      };
      start = this.cursor;
      size = this.readU16();
      zero = this.readU8();
      dataType = this.readU8();
      typedValue.rawType = dataType;
      switch (dataType) {
        case TypedValue.TYPE_INT_DEC:
          typedValue.value = this.readS32();
          typedValue.type = 'int_dec';
          break;
        case TypedValue.TYPE_INT_HEX:
          typedValue.value = this.readS32();
          typedValue.type = 'int_hex';
          break;
        case TypedValue.TYPE_STRING:
          ref = this.readS32();
          typedValue.value = ref > 0 ? this.strings[ref] : '';
          typedValue.type = 'string';
          break;
        case TypedValue.TYPE_REFERENCE:
          id = this.readU32();
          typedValue.value = "resourceId:0x" + (id.toString(16));
          typedValue.type = 'reference';
          break;
        case TypedValue.TYPE_INT_BOOLEAN:
          typedValue.value = this.readS32() !== 0;
          typedValue.type = 'boolean';
          break;
        case TypedValue.TYPE_NULL:
          this.readU32();
          typedValue.value = null;
          typedValue.type = 'null';
          break;
        case TypedValue.TYPE_INT_COLOR_RGB8:
          typedValue.value = this.readHex24();
          typedValue.type = 'rgb8';
          break;
        case TypedValue.TYPE_INT_COLOR_RGB4:
          typedValue.value = this.readHex24();
          typedValue.type = 'rgb4';
          break;
        case TypedValue.TYPE_INT_COLOR_ARGB8:
          typedValue.value = this.readHex32();
          typedValue.type = 'argb8';
          break;
        case TypedValue.TYPE_INT_COLOR_ARGB4:
          typedValue.value = this.readHex32();
          typedValue.type = 'argb4';
          break;
        case TypedValue.TYPE_DIMENSION:
          typedValue.value = this.readDimension();
          typedValue.type = 'dimension';
          break;
        case TypedValue.TYPE_FRACTION:
          typedValue.value = this.readFraction();
          typedValue.type = 'fraction';
          break;
        default:
          type = dataType.toString(16);
          debug("Not sure what to do with typed value of type 0x" + type + ", falling back to reading an uint32");
          typedValue.value = this.readU32();
          typedValue.type = 'unknown';
      }
      end = start + size;
      if (this.cursor !== end) {
        type = dataType.toString(16);
        diff = end - this.cursor;
        debug("Cursor is off by " + diff + " bytes at " + this.cursor + " at supposed end of typed value of type 0x" + type + ". The typed value started at offset " + start + " and is supposed to end at offset " + end + ". Ignoring the rest of the value.");
        this.cursor = end;
      }
      return typedValue;
    };

    ResourcesParser.prototype.convertIntToFloat = function(int) {
      var buf;
      buf = new ArrayBuffer(4);
      new (Int32Array(buf)[0] = buf);
      return new Float32Array(buf)[0];
    };

    ResourcesParser.prototype.readString = function(encoding) {
      var byteLength, stringLength, value;
      switch (encoding) {
        case 'utf-8':
          stringLength = this.readLength8(encoding);
          byteLength = this.readLength8(encoding);
          value = this.buffer.toString(encoding, this.cursor, this.cursor += byteLength);
          this.readU8();
          return value;
        case 'ucs2':
          stringLength = this.readLength16(encoding);
          byteLength = stringLength * 2;
          value = this.buffer.toString(encoding, this.cursor, this.cursor += byteLength);
          this.readU16();
          return value;
        default:
          throw new Error("Unsupported encoding '" + encoding + "'");
      }
    };

    ResourcesParser.prototype.readStringUTF16 = function(strLen) {
      var value = this.buffer.toString("utf8", this.cursor, this.cursor += strLen*2);
      return value;
    }

    ResourcesParser.prototype.readChunkHeader = function() {
      var begin = this.cursor;
      var chunkTypeValue = this.readU16();
      var headerSizeValue = this.readU16();
      var chunkSizeValue = this.readU32();
      switch (chunkTypeValue) {
        case ChunkType.TABLE:
          var resourceTableHeader = {
            chunkType: chunkTypeValue,
            headerSize: headerSizeValue,
            chunkSize: chunkSizeValue,
            packageCount: this.readU32()
          };
          this.cursor = begin + headerSizeValue;
          return resourceTableHeader;
        case ChunkType.STRING_POOL:
          var stringPoolHeader = {
            chunkType: chunkTypeValue,
            headerSize: headerSizeValue,
            chunkSize: chunkSizeValue,
            stringCount: this.readU32(),
            styleCount: this.readU32(),
            flags: this.readU32(),
            stringsStart: this.readU32(),
            stylesStart: this.readU32()
          };
          this.cursor = begin + headerSizeValue;
          return stringPoolHeader;
        case ChunkType.TABLE_PACKAGE:
          var packageHeader = {
            chunkType: chunkTypeValue,
            headerSize: headerSizeValue,
            chunkSize: chunkSizeValue,
            id: this.readU32(),
            name: this.readStringUTF16(128),
            typeStrings: this.readU32(),
            lastPublicType: this.readU32(),
            keyStrings: this.readU32(),
            lastPublicKey: this.readU32()
          };
          this.cursor = begin + headerSizeValue;
          return packageHeader;
        case ChunkType.TABLE_TYPE_SPEC:
          var typeSpecHeader = {
            chunkType: chunkTypeValue,
            headerSize: headerSizeValue,
            chunkSize: chunkSizeValue,
            id: this.readU8(),
            res0: this.readU8(),
            res1: this.readU16(),
            entryCount: this.readU32()
          };
          this.cursor = begin + headerSizeValue;
          return typeSpecHeader;
        case ChunkType.TABLE_TYPE:
          var typeHeader = {
            chunkType: chunkTypeValue,
            headerSize: headerSizeValue,
            chunkSize: chunkSizeValue,
            id: this.readU8(),
            res0: this.readU8(),
            res1: this.readU16(),
            entryCount: this.readU32(),
            entriesStart: this.readU32(),
            config: this.readResTableConfig()
          };
          this.cursor = begin + headerSizeValue;
          return typeHeader;
        case ChunkType.NULL:
          //buffer.skip((int) (chunkSize - headerSize));
        default: 
          throw new Error("Unexpected chunk Type:" + chunkTypeValue);
      }
    };

    function printBuffer(buffer, begin, end) {
      var str = '';
      var a;
      var tmpbf = buffer.slice(begin, end);
      for (var i = 0; i < tmpbf.length; i++) {
        a = tmpbf[i].toString(16);
        if (a.length <= 0) {
          str += "00";
        } else if (a.length == 1) {
          str += ('0' + a);
        } else if (a.length == 2) {
          str += a;
        } else {
          str += "##";
        }
      }
      return str;
    }

    ResourcesParser.prototype.readResTableConfig = function() {
        var begin = this.cursor;
        var size =  this.readU32();
        this.cursor += 4;

        // var lang = this.readU16();
        // var coun = this.readU16();
        this.cursor += 4;
        var language =  this.buffer.toString('ascii', this.cursor-4,  this.cursor-2);
        var country =  this.buffer.toString('ascii', this.cursor-2,  this.cursor);

        // console.log("readResTableConfig %s  %s %s %d %s", printBuffer(this.buffer, begin, begin+size), language, country, size, this.buffer.toString('ascii', begin,  this.cursor));

        var end = this.cursor;
        this.cursor += size - (end - begin);

        return {
          language: language,
          country: country
        };
    }

    ResourcesParser.prototype.readStringPool = function(header) {
      var anchor, encoding, offsets, _i, _j, _ref, _ref1;
      var strings = [];
      // header.stringCount = this.readU32();
      // header.styleCount = this.readU32();
      // header.flags = this.readU32();
      // header.stringsStart = this.readU32();
      // header.stylesStart = this.readU32();
      // console.log('readStringPool header');
      // console.dir(header);
      if (header.chunkType !== ChunkType.STRING_POOL) {
        throw new Error('Invalid string pool header');
      }
      anchor = this.cursor;
      offsets = [];
      for (_i = 0, _ref = header.stringCount; 0 <= _ref ? _i < _ref : _i > _ref; 0 <= _ref ? _i++ : _i--) {
        offsets.push(this.readU32());
      }
      encoding = header.flags & StringFlags.UTF8 ? 'utf-8' : 'ucs2';
      this.cursor = anchor + header.stringsStart - header.headerSize;
      var first_cursor = this.cursor;
      for (_j = 0, _ref1 = header.stringCount; 0 <= _ref1 ? _j < _ref1 : _j > _ref1; 0 <= _ref1 ? _j++ : _j--) {
        // console.log("%d %d %d", _j, this.cursor, offsets[_j], this.cursor - offsets[_j]);
        strings.push(this.readString(encoding));
      }
      this.cursor = anchor + header.chunkSize - header.headerSize;
      return strings;
    };

    ResourcesParser.prototype.readResourceMap = function(header) {
      var count, _i;
      count = Math.floor((header.chunkSize - header.headerSize) / 4);
      for (_i = 0; 0 <= count ? _i < count : _i > count; 0 <= count ? _i++ : _i--) {
        this.resources.push(this.readU32());
      }
      return null;
    };

    ResourcesParser.prototype.readXmlNamespaceStart = function(header) {
      var commentRef, line, prefixRef, uriRef;
      line = this.readU32();
      commentRef = this.readU32();
      prefixRef = this.readS32();
      uriRef = this.readS32();
      return null;
    };

    ResourcesParser.prototype.readXmlNamespaceEnd = function(header) {
      var commentRef, line, prefixRef, uriRef;
      line = this.readU32();
      commentRef = this.readU32();
      prefixRef = this.readS32();
      uriRef = this.readS32();
      return null;
    };

    ResourcesParser.prototype.readXmlElementStart = function(header) {
      var attrCount, attrSize, attrStart, classIndex, commentRef, idIndex, line, nameRef, node, nsRef, styleIndex, _i;
      node = {
        namespaceURI: null,
        nodeType: NodeType.ELEMENT_NODE,
        nodeName: null,
        attributes: [],
        childNodes: []
      };
      line = this.readU32();
      commentRef = this.readU32();
      nsRef = this.readS32();
      nameRef = this.readS32();
      if (nsRef > 0) {
        node.namespaceURI = this.strings[nsRef];
      }
      node.nodeName = this.strings[nameRef];
      attrStart = this.readU16();
      attrSize = this.readU16();
      attrCount = this.readU16();
      idIndex = this.readU16();
      classIndex = this.readU16();
      styleIndex = this.readU16();
      for (_i = 0; 0 <= attrCount ? _i < attrCount : _i > attrCount; 0 <= attrCount ? _i++ : _i--) {
        node.attributes.push(this.readXmlAttribute());
      }
      if (this.document) {
        this.parent.childNodes.push(node);
        this.parent = node;
      } else {
        this.document = this.parent = node;
      }
      this.stack.push(node);
      return node;
    };

    ResourcesParser.prototype.readXmlAttribute = function() {
      var attr, nameRef, nsRef, valueRef;
      attr = {
        namespaceURI: null,
        nodeType: NodeType.ATTRIBUTE_NODE,
        nodeName: null,
        name: null,
        value: null,
        typedValue: null
      };
      nsRef = this.readS32();
      nameRef = this.readS32();
      valueRef = this.readS32();
      if (nsRef > 0) {
        attr.namespaceURI = this.strings[nsRef];
      }
      attr.nodeName = attr.name = this.strings[nameRef];
      if (valueRef > 0) {
        attr.value = this.strings[valueRef];
      }
      attr.typedValue = this.readTypedValue();
      return attr;
    };

    ResourcesParser.prototype.readXmlElementEnd = function(header) {
      var commentRef, line, nameRef, nsRef;
      line = this.readU32();
      commentRef = this.readU32();
      nsRef = this.readS32();
      nameRef = this.readS32();
      this.stack.pop();
      this.parent = this.stack[this.stack.length - 1];
      return null;
    };

    ResourcesParser.prototype.readXmlCData = function(header) {
      var cdata, commentRef, dataRef, line;
      cdata = {
        namespaceURI: null,
        nodeType: NodeType.CDATA_SECTION_NODE,
        nodeName: '#cdata',
        data: null,
        typedValue: null
      };
      line = this.readU32();
      commentRef = this.readU32();
      dataRef = this.readS32();
      if (dataRef > 0) {
        cdata.data = this.strings[dataRef];
      }
      cdata.typedValue = this.readTypedValue();
      this.parent.childNodes.push(cdata);
      return cdata;
    };

    ResourcesParser.prototype.readNull = function(header) {
      this.cursor += header.chunkSize - header.headerSize;
      return null;
    };


    ResourcesParser.prototype.readPackage = function(packageHeader) {
        var pair = {};
        //read packageHeader
        var resourcePackage = {
          name: packageHeader.name,
          id: packageHeader.id
        };

        var beginPos = this.cursor;
        // read type string pool
        if (packageHeader.typeStrings > 0) {
            this.cursor = beginPos + packageHeader.typeStrings
                    - packageHeader.headerSize;
            resourcePackage.typeStringPool = this.readStringPool(this.readChunkHeader());
            // console.log("resourcePackage.typeStringPool %d", resourcePackage.typeStringPool.length);
            // console.dir(resourcePackage.typeStringPool );
        }

        // read key string pool
        if (packageHeader.keyStrings > 0) {
            this.cursor = beginPos + packageHeader.keyStrings
                    - packageHeader.headerSize;
            resourcePackage.keyStringPool = this.readStringPool( this.readChunkHeader());
            // console.log("resourcePackage.keyStringPool  length %d", resourcePackage.keyStringPool.length);
            // console.dir(resourcePackage.keyStringPool );
        }

        resourcePackage.typeSpecMap={};
        resourcePackage.typesMap={};

        var chunkHeader = null;
        // outer:
        var i;
        // console.log("cursor %s length %s", this.cursor, this.buffer.length);
        while (this.cursor < this.buffer.length) {
            // console.log("cursor %s length %s", this.cursor, this.buffer.length);
            chunkHeader = this.readChunkHeader();
            switch (chunkHeader.chunkType) {
                case ChunkType.TABLE_TYPE_SPEC:
                  // console.log("============TABLE_TYPE_SPEC %d", this.cursor);
                   // console.dir(chunkHeader);
                    var typeSpecChunkBegin = this.cursor;
                    var typeSpecHeader = chunkHeader;
                    var entryFlags = new Array(typeSpecHeader.entryCount);
                    for (i = 0; i < typeSpecHeader.entryCount; i++) {
                        entryFlags[i] = this.readU32();
                    }

                    var typeSpec = {
                        id: typeSpecHeader.id
                    }

                    typeSpec.entryFlags=entryFlags;
                    //id start from 1
                    typeSpec.name=resourcePackage.typeStringPool[typeSpecHeader.id - 1];

                    // console.dir(typeSpec);
                    resourcePackage.typeSpecMap[typeSpec.id] = typeSpec;
                    this.cursor = typeSpecChunkBegin + typeSpecHeader.chunkSize - typeSpecHeader.headerSize;
                    break;
                case ChunkType.TABLE_TYPE:
                  // console.log("============TABLE_TYPE %d", this.cursor);
                    var typeChunkBegin = this.cursor;
                    var typeHeader = chunkHeader;
                    // read offsets table
                    var offsets = new Array(typeHeader.entryCount);
                    for (i = 0; i < typeHeader.entryCount; i++) {
                        offsets[i] = this.readU32();
                    }

                    var type = {
                      id: typeHeader.id,
                      locale: {
                        language: typeHeader.config.language,
                        country: typeHeader.config.country
                      }
                    };
                    type.name = resourcePackage.typeStringPool[typeHeader.id - 1];
                    type.offsets = offsets;
                    // console.dir(typeHeader);
                    // console.dir(type);
                    var entryPos = typeChunkBegin + typeHeader.entriesStart
                            - typeHeader.headerSize;
                    this.cursor = entryPos;
                    var b = this.buffer.slice(this.cursor);
                    // b.order(byteOrder);
                    type.buffer = b;
                    type.keyStringPool = resourcePackage.keyStringPool;
                    type.stringPool = this.strings;
                    if (!resourcePackage.typesMap[type.id]){
                      resourcePackage.typesMap[type.id] = [];
                    }
                    resourcePackage.typesMap[type.id].push(type);
                    this.locales.push(type.locale.language + "_" + type.locale.country);
                    this.cursor = typeChunkBegin + typeHeader.chunkSize - typeHeader.headerSize;
                    break;
                case ChunkType.TABLE_PACKAGE:
                  // console.log("============TABLE_PACKAGE %d", this.cursor);
                    // another package. we should read next package here
                    pair.left = resourcePackage;
                    pair.right = chunkHeader;
                    return pair;
                default:
                    throw new Error("unexpected chunk type:" + chunkHeader.getChunkType());
            }
        }

        pair.left = resourcePackage;
        pair.right = chunkHeader;
        return pair;
    }

    function matchLocale(locale, targetLocale){
      // console.log("matchLocale %s %s %s %s", locale.language, locale.country, targetLocale.language, targetLocale.country);
        if (locale == null) {
            return -1;
        }
        if (locale.language == targetLocale.language) {
            if (locale.country == targetLocale.country)  {
                return 4;
            } else if (targetLocale.country == '') {
                return 3;
            } else {
                return 2;
            }
        } else if (targetLocale.language == '') {
            if (locale.country == targetLocale.country)  {
                return 3;
            }
            return 1;
        } else if (locale.language == 'en'){
            return 0;
        } else {
            return -1;
        }
    }

    ResourcesParser.prototype.readResourceEntry = function(type, resourceTable, locale) {
        var beginPos = type.cursor;
        var size = type.buffer.readUInt16LE(type.cursor);
        type.cursor += 2;
        var flags = type.buffer.readUInt16LE(type.cursor);
        type.cursor += 2;
        var resourceEntry = {
          size: size,
          flags: size
        }
        // size is always 8(simple), or 16(complex)
        // resourceEntry.setSize(Buffers.readUShort(buffer));
        // resourceEntry.setFlags(Buffers.readUShort(buffer));
        var keyRef = type.buffer.readUInt32LE(type.cursor);
        resourceEntry.key=type.keyStringPool[keyRef];
        // console.log("keyStringPool lenght %d  keyRef %d key %s", type.keyStringPool.length, keyRef, resourceEntry.key);

        if ((resourceEntry.flags &  0x0001) != 0) {
          console.log("ResourceMapEntry is not implemented");
           return null;
            // ResourceMapEntry resourceMapEntry = new ResourceMapEntry(resourceEntry);

            // // Resource identifier of the parent mapping, or 0 if there is none.
            // resourceMapEntry.setParent(Buffers.readUInt(buffer));
            // resourceMapEntry.setCount(Buffers.readUInt(buffer));

            // buffer.position((int) (beginPos + resourceEntry.getSize()));

            // //An individual complex Resource entry comprises an entry immediately followed by one or more fields.
            // ResourceTableMap[] resourceTableMaps = new ResourceTableMap[(int) resourceMapEntry.getCount()];
            // for (int i = 0; i < resourceMapEntry.getCount(); i++) {
            //     resourceTableMaps[i] = readResourceTableMap();
            // }

            // resourceMapEntry.setResourceTableMaps(resourceTableMaps);
            // return resourceMapEntry;
        } else {
            type.cursor = beginPos + resourceEntry.size;
            resourceEntry.value = this.readResValue(type, resourceTable, locale);
            return resourceEntry;
        }
      }


    ResourcesParser.prototype.getAttrNameById = function(resourceId, locale) {
      return this.getResourceById(resourceId, locale, true);
    }

    ResourcesParser.prototype.getResourceById = function(resourceId, locale, isAttr) {
//        An Android Resource id is a 32-bit integer. It comprises
//        an 8-bit Package id [bits 24-31]
//        an 8-bit Type id [bits 16-23]
//        a 16-bit Entry index [bits 0-15]

        // android system styles.
      if (resourceId > AndroidConstants.SYS_STYLE_ID_START && resourceId < AndroidConstants.SYS_STYLE_ID_END) {
        return "@android:style/" + resourceId.toString(16);
      }

      var str = "resourceId:0x" + resourceId.toString(16);
      if (this.resourceTable == null) {
        return str;
      }

      var packageId = (resourceId >> 24 & 0xff);
      var typeId = ((resourceId >> 16) & 0xff);
      var entryIndex = (resourceId & 0xffff);
      // console.log("getResourceById %s %s %s %s", resourceId.toString(16), packageId.toString(16), typeId.toString(16), entryIndex.toString(16));
      var resourcePackage = this.resourceTable.packageMap[packageId];
      if (resourcePackage == null) {
        // console.log("getResourceById %s is not a valid package.", packageId);
        return str;
      }
      var typeSpec = resourcePackage.typeSpecMap[typeId];
      var types = resourcePackage.typesMap[typeId];
      if (typeSpec == null || types == null) {
        // console.log("getResourceById %s is not a valid type.", typeId);
        return str;
      }
      // console.log("getResourceById typeSpec: %s length %d", typeSpec.name, typeSpec.length);
      // console.log("getResourceById types: %s length %d", types, types.length);
      if (entryIndex >= typeSpec.entryFlags.length) {
        // console.log("getResourceById %s is not a entryIndex.", entryIndex);
        return str;
      }

      // read from type resource
      var resource = null;
      var ref = null;
      var currentLevel = -2;
      var i;
      var arr = locale.split('_');
      var targetLocale = {
        language: arr[0],
        country: arr[1]
      };
      var type;
        for (i = 0; i < types.length; i++) {
          type = types[i];

          //getResourceEntry
          if (entryIndex > type.offsets.length){
            // console.log("getResourceById entryIndex > offsets.length %d %d", entryIndex, offsets.length);
            continue;
          }

        if (type.offsets[entryIndex] == 0xFFFFFFFF) {
            // console.log("type %d offsets[%d] is NO_ENTRY %s %s %s", i, entryIndex, type.offsets[entryIndex].toString(16), type.locale.language, type.locale.country);
            continue;
        }
        // console.log("type %d offsets[%d] is Good %s", i, entryIndex, type.offsets[entryIndex].toString(16));

        type.cursor = type.offsets[entryIndex];
        // console.log("getResourceById resourceEntry cursor %s %s offset %d", this.cursor, entryIndex, type.offsets[entryIndex]);
        var curResourceEntry = this.readResourceEntry(type, this.resourceTable, locale);
        if (curResourceEntry == null) {
          continue;
        }

        // not ResourceMapEntry and resource == 0. some resource entry do not have  effective resource id
        // if (curResourceEntry.value != null && curResourceEntry.value.id == 0 && curResource.getValue().getValue() == null) {
        //   continue;
        // }

        // console.dir(curResourceEntry);
        // console.log("getResourceById curResourceEntry.value %s", curResourceEntry.value);

        ref = curResourceEntry.key;
        var level = matchLocale(type.locale, targetLocale);
        if (level == 4) {
          // console.log("this locale is the best. %s_%s", type.locale.language, type.locale.country);
          resource = curResourceEntry;
          break;
        } else if (level > currentLevel) {
          currentLevel = level;
          resource = curResourceEntry;
          // console.log("this locale is the better. %s_%s", type.locale.language, type.locale.country);
        }
        // console.log("getResourceById level %s %s %s", level, type.locale.language, targetLocale.language);
      }
      var result;
      if (resource == null) {//locale == null || 
        result = "@" + typeSpec.name + "/" + ref;
      } else if (isAttr) {
        result = resource.key;
      } else {
        result = resource.value;
      }
      return result;
    }

    ResourcesParser.prototype.parse = function() {
      var diff, end, header, resMapHeader, start, type, resourceTableHeader;
      resourceTableHeader = this.readChunkHeader();
      // console.dir(resourceTableHeader);
      if (resourceTableHeader.chunkType !== ChunkType.TABLE) {
        throw new Error('Invalid TABLE header');
      }
      this.resourceTable.packageMap = {};
      this.strings = this.readStringPool(this.readChunkHeader());
      // console.dir(this.strings.length);
      packageHeader = this.readChunkHeader();
      // console.log('parse: packageHeader');
      // console.dir(packageHeader);
      for (var i = 0; i < resourceTableHeader.packageCount; i++) {
        var pair = this.readPackage(packageHeader);
        // console.log('parse: packageHeader %d id:%s' , i, packageHeader.id);
        this.resourceTable.packageMap[pair.left.id] = pair.left;
        packageHeader = pair.right;
      }

      // console.log('parse: locales');
      // console.dir(this.locales);
      return this;
    };

    return ResourcesParser;

  })();

  module.exports = ResourcesParser;

}).call(this);
